// lib/cli/index.js

'use strict';

const cwd = process.cwd();
const mask = '**/*.html';

const glob = require('glob').sync;

const {join} = require('path');
const {readFileSync} = require('fs');

const {ErrorCode, OnError} = require('../parser');
const {crossMark, checkMark, arrowNext} = require('./marks');

require('colors');

const createHandler = (filename, {warnings = false, silent = false}) => {
  let errorCount = 0;
  let warningCount = 0;

  const onWarning = (obj) => {
    let message = obj.message;
    const pos = obj.loc ? ('' + obj.loc.line + ':' + obj.loc.column).bold : '';

    if (obj.id === ErrorCode.WARN_LT_IN_TEXT) {
      message = 'Символ «<» в тексте. Замените на «&lt;»'.yellow;
    }

    console.log(
      ` ${crossMark.yellow} ${filename.bold}:${pos} ${message}`
    );
    warningCount++;
  };

  const onWarningSilent = () => {
    warningCount++;
  };

  const onError = (obj) => {
    let message = obj.message;
    const pos = obj.loc ? ('' + obj.loc.line + ':' + obj.loc.column).bold : '';
    let result = OnError.STOP;
    let posOpen;

    if (obj.id === ErrorCode.ERROR_TAG_MISMATCH) {
      message = 'Не совпадают открывающий и закрывающий теги. Закрывающий: '.red +
        `«${obj.close.name}»`.cyan;

      if (obj.open) {
        posOpen = obj.open.loc.line + ':' + obj.open.loc.column;
        message += `\n ${arrowNext.cyan} ${filename.bold}:${posOpen.bold} ` + '… открывающий: '.red +
          `«${obj.open.name}»`.cyan;
      }

      result = OnError.KEEP_MOVING;
    } else if (obj.id === ErrorCode.ERROR_TAG_UNCLOSED) {
      message = 'Тег '.red + `«${obj.open.name}»`.cyan + ' не закрыт'.red;
    } else if (obj.id === ErrorCode.ERROR_CANNOT_PARSE) {
      message = 'Парсер не может справиться с тем, что к нему пришло: '.red +
        `«${obj.tail}»`.cyan;
    } else if (obj.id === ErrorCode.ERROR_LEXER) {
      message = 'Ошибка разбора '.red + obj.message.bold;
    }

    console.log(
      ` ${crossMark.red} ${filename.bold}:${pos} ${message}`
    );

    errorCount++;

    return result;
  };

  const onErrorSilent = () => {
    errorCount++;
  };

  return {
    onError: silent ? onErrorSilent : onError,
    onWarning: warnings && !silent ? onWarning : onWarningSilent,
    results() {
      return {
        errorCount,
        warningCount
      };
    }
  };
};

const ignoredPath = (path) => {
  if (path.startsWith(join(cwd, 'node_modules'))) {
    return true;
  }

  if (path.startsWith(join(cwd, 'release'))) {
    return true;
  }

  return false;
};

module.exports = (argv) => ({
  run() {
    let totalErrors = 0;
    let totalWarnings = 0;

    const opts = {
      warnings: (argv.warnings || argv.w),
      silent: (argv.silent || argv.q)
    };

    let path = argv._[0];

    if (path) {
      path = join(cwd, path);
    } else {
      path = cwd;
    }

    glob(join(path, mask)).forEach((filename) => {
      if (!ignoredPath(filename)) {
        const baseName = filename.substr(cwd.length + 1);

        const createParser = require('../parser').parser;
        const fileContent = readFileSync(filename).toString();
        const parser = createParser(fileContent);

        const handler = createHandler(baseName, opts);

        parser.parse(handler);

        const result = handler.results();

        totalErrors += result.errorCount;
        totalWarnings += result.warningCount;
      }
    });

    if (totalErrors > 0 || totalWarnings > 0) {
      const mark = totalErrors > 0 ? crossMark.red : crossMark.yellow;
      console.log(` ${mark} Ошибок: ${totalErrors}, предупреждений: ${totalWarnings}`);
    } else {
      console.log(` ${checkMark.green} Всё в порядке`);
    }

    if (totalErrors > 0) {
      process.exit(1);
    }
  }
});
