// lib/lexer.js

'use strict';

var Lexem = {
  DOCTYPE: 0,
  TEXT: 1,
  COMMENT: 2,
  TAG: 3,

  // TagType
  OPEN: 4,
  CLOSE: 5
};

exports.Lexem = Lexem;

exports.lexer = function (initialContent) {
  var content = initialContent || '';
  var position = 0;

  var reset = function (newContent) {
    content = newContent;
    this.rewind();
    return this;
  };

  var rewind = function () {
    position = 0;
  };

  var getTail = function () {
    return content.substr(position);
  };

  var endOfText = function (tail) {
    return (
      tail.startsWith('<!') ||
      /^<\/?\w+[^>]*>/.exec(tail)
    );
  };

  var findEndOfText = function () {
    var tail = getTail();
    var pos = tail.indexOf('<');

    while (pos !== -1) {
      var chunk = tail.substr(pos);

      if (endOfText(chunk)) {
        return pos;
      } else {
        pos = tail.indexOf('<', pos + 1);
      }
    }

    return tail.length;
  };

  var fetchChunk = function (chars, fn) {
    var chunk = getTail().substr(0, chars);
    var obj = fn(chars, chunk);
    obj.raw = chunk;
    position += chars;
    return obj;
  };

  var fetchText = function (chars) {
    var tail = getTail();

    if (chars === 0) {
      while (!/<^\w+/.test(tail.substr(chars))) {
        chars = tail.indexOf('<', chars + 1);

        if (chars === -1) {
          chars = tail.length;
          break;
        }
      }
    }

    return fetchChunk(chars, function (_chars, chunk) {
      return {
        type: Lexem.TEXT,
        content: chunk
      };
    });
  };

  var fetchComment = function (chars) {
    return fetchChunk(chars, function (_chars, chunk) {
      return {
        type: Lexem.COMMENT,
        content: chunk. // «<!-- comment -->»
          substr(4). // « comment -->»
          substr(0, _chars - 7). // « comment »
          trim() // «comment»
      };
    });
  };

  var fetchDoctype = function (chars) {
    return fetchChunk(chars, function (_chars, chunk) {
      return {
        type: Lexem.DOCTYPE,
        content: chunk. // «<!DOCTYPE html>»
          substr(9). // « html>»
          substr(0, _chars - 10). // « html>»
          trim() // «html»
      };
    });
  };

  var identifier = /^(\w+)/;
  var attrKey = /^\s+([-\w]+(:[-\w]+)?)\s*=\s*/;
  var emptyAttr = /^\s+(\w+)/;
  var attrValue = /^("([^"]*)"|'([^']*)')/;

  var buildAttributes = function (attrs, error) {
    var lexem = {
      attrs: attrs || {}
    };

    if (typeof error === 'string') {
      lexem.errors = [error];
    }

    return lexem;
  };

  var parseAttributes = function (tail) {
    var attrs = {};
    var md;
    var key;
    var value;

    while (tail.length > 0) {
      md = attrKey.exec(tail);
      if (md) {
        key = md[1];
        tail = tail.substr(md[0].length);

        md = attrValue.exec(tail);

        if (md) {
          value = md[2] || md[3];
          attrs[key] = value;

          tail = tail.substr(md[0].length);
        } else {
          return buildAttributes(
            attrs,
            'Cannot find attribute value «' + tail + '»'
          );
        }
      } else {
        md = emptyAttr.exec(tail);

        if (md) {
          key = md[1];
          tail = tail.substr(md[0].length);

          attrs[key] = null;
        } else {
          if (tail.trim() === '') {
            tail = '';
          } else {
            return buildAttributes(
              attrs,
              'Cannot find attribute name «' + tail + '»'
            );
          }
        }
      }
    }

    return buildAttributes(attrs);
  };

  var buildTag = function (name, close, forceEmpty, tail, error) {
    var lexem = {
      name: name,
      type: Lexem.TAG,
      tagType: close ? Lexem.CLOSE : Lexem.OPEN,
      forceEmpty: forceEmpty
    };

    if (!close) {
      var attrs = parseAttributes(tail);
      Object.assign(lexem, attrs);
    }

    if (typeof error === 'string') {
      if (!lexem.errors) {
        lexem.errors = [];
      }
      lexem.errors.push(error);
    }

    return lexem;
  };

  var parseTagChunk = function (chunk) {
    var tail = chunk;
    var close = tail.startsWith('</');
    var forceEmpty = tail.endsWith('/>');
    tail = tail.substr(close ? 2 : 1);
    tail = tail.substr(0, tail.length - (forceEmpty ? 2 : 1));

    var md = identifier.exec(tail);
    var name = null;
    var error = null;

    if (md) {
      name = md[1];

      tail = tail.substr(md[0].length);
    } else {
      error = 'Tag name not found: «' + tail + '», «' + chunk + '»';
    }

    return buildTag(
      name.toLowerCase(),
      close,
      forceEmpty,
      tail,
      error
    );
  };

  var fetchTag = function (chars) {
    return fetchChunk(chars, function (_chars, chunk) {
      return Object.assign({
        type: Lexem.TAG
      }, parseTagChunk(chunk));
    });
  };

  var fetchLexem = function () {
    var tail = getTail();

    if (tail.length === 0) {
      return null;
    }

    var laIndex = findEndOfText();
    var raIndex;

    if (laIndex > 0) {
      return fetchText(laIndex);
    } else {
      var doctypeConst = '<!doctype';
      var doctypeHead = tail.substr(0, doctypeConst.length).toLowerCase();

      if (doctypeHead === doctypeConst) {
        raIndex = tail.indexOf('>');

        return fetchDoctype(raIndex + 1);
      } else if (tail.startsWith('<!--')) {
        var rcIndex = tail.indexOf('-->');

        if (rcIndex === -1) {
          throw new Error('Unclosed comment: «' + tail + '»');
        } else {
          return fetchComment(rcIndex + 3);
        }
      } else {
        raIndex = tail.indexOf('>');

        return fetchTag(raIndex + 1);
      }
    }
  };

  return {
    reset: reset,
    tail: getTail,
    fetch: fetchLexem,
    rewind: rewind
  };
};
