// lib/parser.js

'use strict';

var lx = require('./lexer');
var lexer = lx.lexer;
var Lexem = lx.Lexem;

var createLocation = require('./location');

var OnError = {
  KEEP_MOVING: 1,
  STOP: 2
};

var ErrorCode = {
  ERROR_TAG_MISMATCH: 1,
  ERROR_TAG_UNCLOSED: 2,
  ERROR_CANNOT_PARSE: 3,
  ERROR_LEXER: 4,

  WARN_LT_IN_TEXT: 101
};

var ParserMethods = [
  'onDoctype',
  'onText',
  'onComment',
  'onOpenTag',
  'onCloseTag',
  'onError',
  'onWarning'
];

exports.OnError = OnError;

exports.ParserMethods = ParserMethods;

exports.ErrorCode = ErrorCode;

exports.parser = function (initialContent) {
  var sink = lexer(initialContent);

  /**
   * Сменить текущее содержимое лексера
   *
   * @param {string} newContent
   * @return {object}
   */
  var reset = function (newContent) {
    sink.reset(newContent);

    return this;
  };

  /**
   * Вызвать метод обработчика
   *
   * @param {object} handler
   * @param {string} method
   * @param {object} lexem
   *
   * @return {any}
   */
  var callMethod = function (handler, method, lexem) {
    var result = null;
    if (typeof handler[method] === 'function') {
      result = handler[method](lexem);
    }
    return result;
  };

  /**
   * Совпадают ли теги
   *
   * @param {object} close
   * @param {object} open
   *
   * @return {boolean}
   */
  var matchesOpenTag = function (close, open) {
    return (
      open &&
      open.name === close.name &&
      open.type === Lexem.TAG &&
      open.tagType === Lexem.OPEN
    );
  };

  var closePrevious = [
    'li'
  ];

  var emptyTagNames = require('./empty-tags');

  /**
   * Должен ли тег быть пустым
   *
   * @param {object} lexem
   * @return {boolean}
   */
  var tagShouldBeEmpty = function (lexem) {
    return (
      lexem.type === Lexem.TAG && (
        emptyTagNames.indexOf(lexem.name) > -1 ||
        lexem.forceEmpty
      )
    );
  };

  /**
   * Распарсить html-строку
   *
   * @param {object} handler
   */
  var parse = function (handler) {
    sink.rewind();

    var lexem;
    var open;
    var broken = false;
    var keepMoving = handler.keepMoving;

    var stack = [];

    var location = createLocation();

    var shouldClosePrevious = function (openLexem) {
      if (handler.ignoreMismatches) {
        return false;
      }

      var last = stack[stack.length - 1];

      return (
        last && last.name === openLexem.name &&
        closePrevious.indexOf(openLexem.name) > -1
      );
    };

    var onError = function (obj) {
      var result = callMethod(handler, 'onError', obj);
      var gotBroken = false;

      if (!keepMoving) {
        if (result !== OnError.KEEP_MOVING) {
          gotBroken = true;
        }
      }

      return gotBroken;
    };

    var recoverCloseTag = function (openTag, close) {
      // <div><tag><h1>Title</div>
      //
      // stack [div, tag, h1]
      // open h1
      // close div
      //
      // TODO: </h1></tag></div>

      if (!openTag || !openTag.name) {
        return false;
      }

      var idx = stack.length - 1;

      for (; idx >= 0; idx--) {
        if (stack[idx].name === close.name) {
          break;
        }
      }

      callMethod(handler, 'onCloseTag', {
        type: Lexem.TAG,
        tagType: Lexem.CLOSE_TAG,
        name: openTag.name,
        raw: ''
      });

      if (idx > -1) {
        for (var i = idx; i < stack.length; i++) {
          callMethod(handler, 'onCloseTag', {
            type: Lexem.TAG,
            tagType: Lexem.CLOSE_TAG,
            name: stack[i].name,
            raw: ''
          });
        }

        stack = stack.slice(0, idx);

        return true;
      } else {
        return false;
      }
    };

    var fetchLexem = function () {
      var result = null;

      try {
        result = sink.fetch();

        if (result) {
          result.loc = location.get();
          location.pass(result.raw);
        }
      } catch (err) {
        callMethod(handler, 'onError', {
          id: ErrorCode.ERROR_LEXER,
          message: err.message,
          loc: location.get()
        });
      }

      if (result && result.errors) {
        result.errors.forEach(function (message) {
          callMethod(handler, 'onError', {
            id: ErrorCode.ERROR_LEXER,
            message: message,
            loc: location.get()
          });
        });
      }

      return result;
    };

    lexem = fetchLexem();

    while (lexem && !broken) {
      lexem.parent = stack.length === 0 ? null : stack[stack.length - 1];

      callMethod(handler, 'onLexem', lexem);

      switch (lexem.type) {
      case Lexem.DOCTYPE:
        callMethod(handler, 'onDoctype', lexem);
        break;

      case Lexem.TAG:
        if (lexem.tagType === Lexem.OPEN) {
          if (shouldClosePrevious(lexem)) {
            open = stack.pop();

            callMethod(handler, 'onError', {
              id: ErrorCode.ERROR_TAG_UNCLOSED,
              message: 'Tag is not closed',
              open: open,
              loc: lexem.loc
            });

            callMethod(handler, 'onCloseTag', {
              type: Lexem.TAG,
              tagType: Lexem.CLOSE,
              name: open.name,
              raw: ''
            });
          }

          if (!tagShouldBeEmpty(lexem)) {
            stack.push(lexem);
          }

          callMethod(handler, 'onOpenTag', lexem);
        } else {
          open = stack.pop();

          if (matchesOpenTag(lexem, open)) {
            lexem.open = open;
            callMethod(handler, 'onCloseTag', lexem);
          } else if (handler.ignoreMismatches) {
            callMethod(handler, 'onCloseTag', lexem);
          } else {
            broken = onError({
              id: ErrorCode.ERROR_TAG_MISMATCH,
              loc: lexem.loc,
              message: 'Unmatched tag',
              open: open,
              close: lexem
            });

            if (!broken) {
              /* broken = ! */ recoverCloseTag(open, lexem);
            }
          }
        }

        break;

      case Lexem.COMMENT:
        callMethod(handler, 'onComment', lexem);
        break;

      case Lexem.TEXT:
        if (lexem.content.includes('<')) {
          callMethod(handler, 'onWarning', {
            id: ErrorCode.WARN_LT_IN_TEXT,
            loc: lexem.loc,
            message: 'Invalid character in text content',
            lexem: lexem
          });
        }

        callMethod(handler, 'onText', lexem);
        break;

      default:
        broken = onError({
          id: ErrorCode.ERROR_CANNOT_PARSE,
          message: 'Cannot parse input',
          tail: sink.tail().substr(0, 80) + '...'
        });
        break;
      }

      lexem = fetchLexem();
    }
  };

  return {
    reset: reset,
    parse: parse
  };
};
