// test/lexer-test.js

'use strict';

const {expect} = require('chai');
const {lexer, Lexem} = require('../lib/lexer');

describe('lexer', () => {
  let lr;
  let lexem;

  beforeEach(() => {
    lr = lexer();
  });

  context('start with doctype', () => {
    context('in uppercase', () => {
      beforeEach(() => {
        lexem = lr.reset('<!DOCTYPE html> foobar').fetch();
      });

      it('should return doctype lexem', () => {
        expect(lexem.type).to.eq(Lexem.DOCTYPE);
      });
    });

    context('in lowercase', () => {
      beforeEach(() => {
        lexem = lr.reset('<!doctype html> foobar').fetch();
      });

      it('should return doctype lexem', () => {
        expect(lexem.type).to.eq(Lexem.DOCTYPE);
      });
    });
  });

  context('start with comment', () => {
    beforeEach(() => {
      lexem = lr.reset('<!-- comment --> foobar').fetch();
    });

    it('should return comment lexem', () => {
      expect(lexem.type).to.eq(Lexem.COMMENT);
      expect(lexem.content).to.eq('comment');
    });

    it('should not contain angle brackets', () => {
      expect(lexem.content).not.to.include('<');
      expect(lexem.content).not.to.include('>');
    });

    it('should keep tail', () => {
      expect(lr.tail()).to.eq(' foobar');
    });
  });

  context('start with comment with «>» inside', () => {
    beforeEach(() => {
      lexem = lr.reset('<!-- life > fear --> foobar').fetch();
    });

    it('should return comment lexem', () => {
      expect(lexem.type).to.eq(Lexem.COMMENT);
      expect(lexem.content).to.eq('life > fear');
    });
  });

  context('start with text', () => {
    beforeEach(() => {
      lexem = lr.reset('some text before <tag/>').fetch();
    });

    it('should return text lexem', () => {
      expect(lexem.type).to.eq(Lexem.TEXT);
      expect(lexem.content).to.eq('some text before ');
    });
  });

  context('start with text with «<» inside', () => {
    beforeEach(() => {
      lexem = lr.reset('four <= five <tag>').fetch();
    });

    it('should return text lexem', () => {
      expect(lexem.type).to.eq(Lexem.TEXT);
      expect(lexem.raw).to.eq('four <= five ');
    });
  });

  context('start with newline', () => {
    beforeEach(() => {
      lexem = lr.reset('\n<html>').fetch();
    });

    it('should return text lexem', () => {
      expect(lexem).to.be.an('object');
      expect(lexem.type).to.eq(Lexem.TEXT);
      expect(lexem.content).to.eq('\n');
    });
  });

  context('start with open tag', () => {
    beforeEach(() => {
      lexem = lr.reset('<a href="#" class="foo_bar" xlink:href="bar">Link</a>').fetch();
    });

    it('should return tag lexem', () => {
      expect(lexem.type).to.eq(Lexem.TAG);
      expect(lexem.tagType).to.eq(Lexem.OPEN);
    });

    it('should have correct tag name', () => {
      expect(lexem.name).to.eq('a');
    });

    it('should have attributes parsed', () => {
      expect(lexem.attrs.href).to.eq('#');
      expect(lexem.attrs.class).to.eq('foo_bar');
      expect(lexem.attrs['xlink:href']).to.eq('bar');
    });
  });

  context('start with mix-cased open tag', () => {
    beforeEach(() => {
      lexem = lr.reset('<stROng>Vodka').fetch();
    });

    it('should lowercase tag name', () => {
      expect(lexem.name).to.eq('strong');
    });
  });

  context('start with open tag with spaces after attributes', () => {
    beforeEach(() => {
      lexem = lr.reset('<a href="#" >Link</a>').fetch();
    });

    it('should return tag lexem', () => {
      expect(lexem.type).to.eq(Lexem.TAG);
      expect(lexem.tagType).to.eq(Lexem.OPEN);
    });
  });

  context('start with close tag', () => {
    beforeEach(() => {
      lexem = lr.reset('</b> foo').fetch();
    });

    it('should return close tag lexem', () => {
      expect(lexem.type).to.eq(Lexem.TAG);
      expect(lexem.tagType).to.eq(Lexem.CLOSE);
    });

    it('should update position', () => {
      expect(lr.tail()).to.eq(' foo');
    });
  });

  context('start with mix-cased close tag', () => {
    beforeEach(() => {
      lexem = lr.reset('</stROng>will').fetch();
    });

    it('should lowercase tag name', () => {
      expect(lexem.type).to.eq(Lexem.TAG);
      expect(lexem.tagType).to.eq(Lexem.CLOSE);
      expect(lexem.name).to.eq('strong');
    });
  });

  context('start with empty buffer', () => {
    beforeEach(() => {
      lexem = lr.fetch();
    });

    it('should return null', () => {
      expect(lexem).to.be.a('null');
    });
  });
});
