// test/parser-test.js

'use strict';

const {parser} = require('../lib/parser');
const {expect} = require('chai');

const handler = () => ({
  broken: false,

  result: [],

  onOpenTag(obj) {
    this.result.push({onOpenTag: obj});
  },
  onCloseTag(obj) {
    this.result.push({onCloseTag: obj});
  },
  onText(obj) {
    this.result.push({onText: obj});
  },
  onError(obj) {
    this.result.push({onError: obj});
    this.broken = true;
  }
});

const openTag = (name) => (evt) => {
  const lexem = evt.onOpenTag;

  return (
    lexem &&
    lexem.name === name
  );
};

const closeTag = (name) => (evt) => {
  const lexem = evt.onCloseTag;

  return (
    lexem &&
    lexem.name === name
  );
};

describe('parser()', () => {
  let pr;
  let hn;
  let result;

  beforeEach(() => {
    pr = parser();
    hn = handler();
  });

  context('simple open-close tags', () => {
    beforeEach(() => {
      pr.reset('<b></b>').parse(hn);
      result = hn.result;
    });

    it('should get tags info', () => {
      expect(result.length).to.eq(2);
      expect(result[0].onOpenTag.name).to.eq('b');
      expect(result[1].onCloseTag.name).to.eq('b');
    });

    it('should get parent ref to null', () => {
      expect(result[0].onOpenTag.parent).to.be.a('null');
    });
  });

  context('unmatched open-close tags', () => {
    beforeEach(() => {
      pr.reset('<b></xb>').parse(hn);
      result = hn.result;
    });

    it('should get error', () => {
      expect(result.length).to.eq(2);
      expect(hn.broken).to.eq(true);
      expect(result[1].onError.message).to.eq('Unmatched tag');
      expect(result[1].onError.open.name).to.eq('b');
      expect(result[1].onError.close.name).to.eq('xb');
    });
  });

  context('text inside tags', () => {
    beforeEach(() => {
      pr.reset('<b>Bold</b>').parse(hn);
      result = hn.result;
    });

    it('should get three lexems', () => {
      expect(result.length).to.eq(3);
      expect(hn.broken).to.eq(false);
      expect(result[1].onText.content).to.eq('Bold');
      expect(result[1].onText.parent.name).to.eq('b');
    });
  });

  context('two img tags', () => {
    beforeEach(() => {
      pr.reset('<img><img>').parse(hn);
      result = hn.result;
    });

    it('should get two lexems', () => {
      expect(result.length).to.eq(2);
      expect(hn.broken).to.eq(false);
    });

    it('should not make first lexem as parent to second', () => {
      expect(result[1].onOpenTag.parent).to.be.a('null');
    });
  });

  context('two self-closed tags', () => {
    beforeEach(() => {
      pr.reset('<foo/><bar/>').parse(hn);
      result = hn.result;
    });

    it('should get two lexems', () => {
      expect(result.length).to.eq(2);
      expect(hn.broken).to.eq(false);
    });

    it('should not make first lexem as parent to second', () => {
      expect(result[1].onOpenTag.parent).to.be.a('null');
    });
  });

  context('error tag, then script', () => {
    let iOpenTag;
    let scriptOpenTag;
    let scriptCloseTag;

    beforeEach(() => {
      pr.reset('<i foo ;;;><script src="script.js"></script>').parse(hn);
      result = hn.result;

      iOpenTag = result.find(openTag('i'));
      scriptOpenTag = result.find(openTag('script'));
      scriptCloseTag = result.find(closeTag('script'));
    });

    it('should find open i tag', () => {
      expect(iOpenTag).to.be.an('object');
    });

    it('should find open script tag', () => {
      expect(scriptOpenTag).to.be.an('object');
    });

    it('should find close script tag', () => {
      expect(scriptCloseTag).to.be.an('object');
    });
  });
});
