// test/transformer-test.js

'use strict';

const {parser, ParserMethods} = require('../lib/parser');
const {transformer} = require('../lib/transformer');
const {expect} = require('chai');

describe('transformer()', () => {
  it('should be a function', () => {
    expect(transformer).to.be.a('function');
  });

  context('transformer by default', () => {
    let tr;

    beforeEach(() => {
      tr = transformer();
    });

    it('should be an object', () => {
      expect(tr).to.be.an('object');
    });

    it('should be a parser\'s handler', () => {
      ParserMethods.forEach((key) => {
        expect(tr).to.respondTo(key);
      });
    });

    context('transforming', () => {
      let pr;
      let result;

      beforeEach(() => {
        pr = parser('<b>Hello.</b>');
        pr.parse(tr);
        result = tr.output();
      });

      it('should return same content', () => {
        expect(result).to.eq('<b>Hello.</b>');
      });
    });
  });

  context('transform that replaces link tag', () => {
    let tr;
    let pr;
    let result;

    beforeEach(() => {
      tr = transformer({
        onOpenTag: function (lexem) {
          if (lexem.name.toLowerCase() === 'link' && lexem.attrs.href === 'style.css') {
            this.push(
              this.tag('style', 'h1 { color: red; }')
            );
          } else {
            this.pass(lexem);
          }
        }
      });

      pr = parser('<head><link href="style.css"></head>');
      pr.parse(tr);

      result = tr.output();
    });

    it('should replace link to style tag', () => {
      expect(result).to.eq('<head><style>h1 { color: red; }</style></head>');
    });
  });

  context('transform that replaces script tag', () => {
    let tr;
    let pr;
    let result;

    const placeHolderScriptTag = (lexem) => {
      const attrs = lexem.attrs || {};

      return (
        lexem.name === 'script' &&
        attrs.src === 'script.js'
      );
    };

    beforeEach(() => {
      tr = transformer({
        onOpenTag: function (lexem) {
          if (placeHolderScriptTag(lexem)) {
            this.push(
              this.tag('script', 'console.log("Hello, transformer");')
            );
          } else {
            this.pass(lexem);
          }
        },
        onCloseTag: function (lexem) {
          if (!placeHolderScriptTag(lexem.open)) {
            this.pass(lexem);
          }
        }
      });

      pr = parser('<body><p>Drink that coffee</p><script src="script.js"></script></body>');
      pr.parse(tr);

      result = tr.output();
    });

    it('should replace script tag with script content', () => {
      expect(result).to.eq('<body><p>Drink that coffee</p><script>console.log("Hello, transformer");</script></body>');
    });
  });

  context('transform that removes html comments', () => {
    let tr;
    let pr;
    let result;

    beforeEach(() => {
      tr = transformer({
        onComment: function () {} // just skip
      });

      pr = parser('<body><!-- <p>Drink that coffee</p><script src="script.js"></script> --></body>');
      pr.parse(tr);

      result = tr.output();
    });

    it('should return empty body', () => {
      expect(result).to.eq('<body></body>');
    });
  });

  context('transform that removes spaces', () => {
    let tr;
    let pr;
    let result;

    const hasPreParent = (lexem) => {
      const parent = lexem.parent;

      return (
        parent &&
        (parent.name === 'code' || hasPreParent(parent))
      );
    };

    beforeEach(() => {
      tr = transformer({
        onText: function (lexem) {
          if (hasPreParent(lexem)) {
            this.pass(lexem);
          } else {
            this.push(lexem.content.replace(/\s+/g, ' '));
          }
        }
      });

      pr = parser('<p>   Hello,     everybody!    </p><code>   <i>1  </i>, 2  </code>');
      pr.parse(tr);

      result = tr.output();
    });

    it('should remove extra spaces', () => {
      expect(result).to.eq('<p> Hello, everybody! </p><code>   <i>1  </i>, 2  </code>');
    });
  });

  context('transform than does not change anything and code with code after </html>', () => {
    let tr;
    let pr;
    let result;

    beforeEach(() => {
      tr = transformer({});

      pr = parser('<html><body></body></html><p>Hello, keep urslf alive!</p></form>');
    });

    it('should not fail', () => {
      expect(
        function () {
          pr.parse(tr);
        }
      ).to.not.throw();
    });

    it('should return code after </html>', () => {
      pr.parse(tr);
      result = tr.output();

      expect(result).to.include('<p>Hello, keep urslf alive!</p>');
    });
  });

  context('transform that ignores mismatches', () => {
    let tr;
    let pr;
    let result;

    beforeEach(() => {
      tr = transformer({ignoreMismatches: true});
      pr = parser('<html><body><table><tr></tr></tr></tr></table></body></html>');
    });

    it('should return all the code', () => {
      pr.parse(tr);
      result = tr.output();

      expect(result).to.eq('<html><body><table><tr></tr></tr></tr></table></body></html>');
    });
  });
});
